package me.hammerle.code;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Stack;
import me.hammerle.exceptions.NoSuchMethodException;
import me.hammerle.exceptions.PreScriptException;
import me.hammerle.math.Fraction;

public class LineCompiler 
{
    private int realLine;
    // counts brackets to know if we are in a function
    private int bracketCounter;
    // last special character to stop push
    private final Stack<Integer> lastSpecial;
    // comma counter, for functions
    private final Stack<Integer> commaCounter;
    // store for adding function on next successfull push
    private final HashMap<Integer, String> nextFunction;
    // stores if a function was just called
    private boolean function;
    
    private String line;
    private byte layer;
    
    private SnuviParser parser;
    private String scriptName;
    
    public LineCompiler(SnuviParser parser, String scriptName)
    {
        realLine = 0;
        bracketCounter = 0;
        lastSpecial = new Stack<>();
        commaCounter = new Stack<>();
        nextFunction = new HashMap<>();
        function = false;
        line = "";
        layer = 0;
        this.parser = parser;
        this.scriptName = scriptName;
    }

    public void reset(int realLine, byte layer, String line) 
    {
        this.realLine = realLine;
        bracketCounter = 0;
        lastSpecial.clear();
        lastSpecial.push(line.length());
        commaCounter.clear();
        nextFunction.clear();
        function = false;
        this.line = line;
        this.layer = layer;
    }
    
    private boolean push(HashMap<String, String> strings, ArrayList<Code> co, int index)
    {
        if(function)
        {
            //System.out.println("push last result");
            if(index + 1 != lastSpecial.peek())
            {
                co.add(new Code(Code.convertInput(strings, line.substring(index + 1, lastSpecial.pop()), true), realLine, layer));
            }
        }
        else
        {
            co.add(new Code(Code.convertInput(strings, line.substring(index + 1, lastSpecial.pop()), true), realLine, layer));
        }
        String sNextFunction = nextFunction.remove(bracketCounter);
        if(sNextFunction != null)
        {
            if(sNextFunction.equals("sub") && !co.isEmpty())
            {
                Code c = co.get(co.size() - 1);
                if("".equals(c.value))
                {
                    c.value = new Fraction(0);
                }
            }
            co.add(new Code(sNextFunction, 2, realLine, layer));
            return true;
        }
        return false;
    }
    
    private boolean isAllowedChar(char c)
    {
        return Character.isLetterOrDigit(c) || c == '.' || c == '_';
    }
    
    public void compile(ArrayList<Code> co, HashMap<String, String> strings)
    {
        int old;
        int j = line.length() - 1;    
        int commas;
        String syntax;
        while(j >= 0)
        {
            syntax = null;
            switch(line.charAt(j))
            {
                case '+': syntax = "add"; break;
                case '-': syntax = "sub"; break;
                case '*': syntax = "mul"; break;
                case '/': syntax = "div"; break;
                case '%': syntax = "math.mod"; break;
                case '^': syntax = "math.pow"; break;
                case '<': syntax = "less"; break;
                case '>': syntax = "greater"; break;
                /* 
                    changeBiSyntax("||", "or", sb, false);
                    changeBiSyntax("&&", "and", sb, false);
                */  
                case '=':
                {
                    char before = line.charAt(j - 1);
                    switch(before)
                    {
                        case '=': syntax = "equal"; break;
                        case '!': syntax = "notequal"; break;
                        case '>': syntax = "greaterequal"; break;
                        case '<': syntax = "lessequal"; break;
                    }
                    if(syntax != null)
                    {
                        function = push(strings, co, j);
                        nextFunction.put(bracketCounter, syntax);
                        j--;
                        lastSpecial.push(j);
                        j--;
                        continue;
                    }
                    function = push(strings, co, j);
                    //if(function)
                    //{
                        //System.out.println("push last result");
                    //}
                    int checker = j - 2;
                    while(checker >= 0)
                    {
                        if(!isAllowedChar(line.charAt(checker)))
                        {
                            throw new PreScriptException(scriptName, realLine, "unexpected character");
                        }
                        checker--;
                    }
                    String var;
                    switch(before)
                    {
                        case '+': syntax = "add"; break;
                        case '-': syntax = "sub"; break;
                        case '*': syntax = "mul"; break;
                        case '/': syntax = "div"; break;
                        default:
                            var = line.substring(0, j);
                            co.add(new Code(var, realLine, layer));
                    }
                    if(syntax != null)
                    {
                        var = line.substring(0, j - 1);
                        co.add(new Code(Code.convertInput(strings, var, true), realLine, layer));
                        co.add(new Code(syntax, 2, realLine, layer));
                        //System.out.println("push last result");
                        co.add(new Code(var, realLine, layer));
                    }
                    co.add(new Code("setvar", 2, realLine, layer));
                    j = -1;
                    continue;
                }
                case ')':
                {
                    bracketCounter++;
                    lastSpecial.push(j);
                    commaCounter.push(0);
                    break;
                }
                case ',':
                {
                    commaCounter.push(commaCounter.pop() + 1);
                    function = push(strings, co, j);
                    lastSpecial.push(j);
                    break;
                }
                case '(':
                {
                    commas = commaCounter.pop() + 1;
                    //if(function)
                    //{
                        //System.out.println("push last result");
                    //}
                    if(line.charAt(j + 1) != ')')
                    {
                        if(j + 1 != lastSpecial.peek())
                        {
                            co.add(new Code(Code.convertInput(strings, line.substring(j + 1, lastSpecial.pop()), true), realLine, layer));
                        }
                    }
                    else
                    {
                        commas = 0;
                    }
                    String sNextFunction = nextFunction.remove(bracketCounter);
                    if(sNextFunction != null)
                    {
                        co.add(new Code(sNextFunction, 2, realLine, layer));
                        //System.out.println("push last result");
                    }
                    bracketCounter--;
                    function = true;
                    old = j;
                    j--;
                    while(j >= 0 && isAllowedChar(line.charAt(j)))
                    {
                        j--;
                    }
                    j++;
                    String functionName = line.substring(j, old);
                    if(!parser.isRegisteredFunction(functionName))
                    {
                        throw new NoSuchMethodException(scriptName, realLine, functionName);
                    }
                    co.add(new Code(functionName, commas, realLine, layer));
                    lastSpecial.push(j);
                }
            }
            if(syntax != null)
            {
                function = push(strings, co, j);
                nextFunction.put(bracketCounter, syntax);
                lastSpecial.push(j);
            }
            j--;
        }
    }
}